
module Dictionary
:
sig
	type ('a,'b) t
	val empty : ('a,'b) t
	val add   : 'a -> 'b -> ('a,'b) t -> ('a,'b) t
	val find  : 'a -> ('a,'b) t -> 'b 
	val insert  : 'a -> 'b -> ('b -> unit) -> ('a,'b) t -> ('a,'b) t
	val member  : 'a -> ('b -> bool) -> ('a,'b) t -> bool 
	val iterate : 'a -> ('b -> unit) -> ('a,'b) t -> unit 
end
=
struct
	type ('a,'b) t =
		('a,'b) dictionary option
	and ('a,'b) dictionary =
		{
		mutable left: ('a,'b) t;
		key: 'a; item: 'b;
		mutable right: ('a,'b) t;
		}
	let empty =
		(None : ('a,'b) t)					
	let rec add k x t =
		match t with 
		| None -> Some {left=None;key=k;item=x;right=None} 
		| Some n ->
			( match compare k n.key with 
			| c when c < 0 -> n.left  <- add k x n.left 
			| c when c > 0 -> n.right <- add k x n.right 
			| _ -> failwith "duplicate identifier" ); 
			t
	let rec find k = function
		| None -> raise Not_found
		| Some n ->
			match compare k n.key with 
			| c when c < 0 -> find k n.left  
			| c when c > 0 -> find k n.right
			| _ -> n.item 
	let rec insert k x f t =
		match t with 
		| None -> Some {left=None;key=k;item=x;right=None} 
		| Some n ->
			( match compare k n.key with 
			| c when c < 0 -> n.left  <- insert k x f n.left 
			| c when c > 0 -> n.right <- insert k x f n.right 
			| _ -> f n.item ); 
			t
	let rec member k p = function
		| None -> false
		| Some n ->
			match compare k n.key with 
			| c when c < 0 -> member k p n.left  
			| c when c > 0 -> member k p n.right
			| _ -> p n.item 
	let rec iterate k f = function
		| None -> ()
		| Some n ->
			match compare k n.key with 
			| c when c < 0 -> iterate k f n.left  
			| c when c > 0 -> iterate k f n.right
			| _ -> f n.item 
end
 
module BiMap
:
sig
	type 'a t
	val singleton : 'a -> 'a -> 'a t
	val insert : 'a -> 'a -> 'a t -> unit
	val member : 'a -> 'a -> 'a t -> bool
	val iterate: ('a -> 'a -> unit) -> 'a t -> unit 
	val iterate_fst : 'a -> ('a -> unit) -> 'a t -> unit 
	val iterate_snd : 'a -> ('a -> unit) -> 'a t -> unit 
end
=
struct
	type 'a t =
		{
		mutable left: 'a edges;
		x: 'a; y: 'a;
		mutable right:'a edges;
		}	
	and 'a edges =
		'a t option
	let singleton x y =
		{left=None;x;y;right=None} 
	let rec insert a b = function 
		| None -> singleton a b 
		| Some n ->
			let c = compare a n.x in
			if c < 0 then n.left <- Some (insert b a n.left)
			else if c > 0 then n.right <- Some (insert b a n.right)
			else if b <> n.y then n.right <- Some (insert b a n.right);
			n
	let insert a b t =
		ignore(insert a b (Some t))
	let rec member a b = function 
		| None -> false
		| Some {left;x;y;right} ->
			(a = x && b = y)
			|| member b a (if a < x then left else right)		
	let member a b t =
		member a b (Some t)
	let flip f x y =
		f y x
	let rec iterate f g = function
		| None -> ()
		| Some {left;x;y;right} -> iterate g f left; f x y; iterate g f right
	let iterate f t =
		iterate f (flip f) (Some t) 
	let rec iterate_1 a f = function
		| None -> ()
		| Some {left;x;y;right} ->
			match compare a x with
			| c when c < 0 -> iterate_2 a f left 
			| c when c > 0 -> iterate_2 a f right
			| _ -> f y; iterate_2 a f right
	and iterate_2 b f = function
		| None -> ()
		| Some {left;x;y;right} ->
			iterate_1 b f left;
			if b = y then f x;
			iterate_1 b f right
	let iterate_snd a f t =
		iterate_1 a f (Some t)
	let iterate_fst b f t =
		iterate_2 b f (Some t)
end

module type DiGraph
=
sig
	type ('a,'b) t
	val make : unit -> ('a,'b) t
	val insert : 'a -> 'b -> 'a -> ('a,'b) t -> unit
	val member : 'a -> 'b -> 'a -> ('a,'b) t -> bool
	val find_from_to : 'b -> ('a -> 'a -> unit) -> ('a,'b) t -> unit 
	val find_from    : ('a -> unit) -> 'b -> 'a -> ('a,'b) t -> unit
	val find_to      : 'a -> 'b -> ('a -> unit) -> ('a,'b) t -> unit
	val marshal_flags : Marshal.extern_flags list
end

module type ER_DiGraph
=
sig
	type t
	type name =
		string
	type relation =
		name
	type concept =
		| Concept  of name
		| ConceptN of name * name
		| ConceptI of name * int
		| ConceptS of name * string 
		| ConceptD of marked (* marker definition *)
		| ConceptU of marked (* marker usage *)
		| ConceptV of marked (* variable definition *)
	and marked =
		{mutable concept:concept; marker:name} 
	type pattern =
		(concept * relation * concept) list
	val make : unit -> t
	val insert  : concept -> relation -> concept -> t -> unit
	val select  : pattern -> t -> unit
	val success : (unit -> unit) ref
	val empty : concept
	val marshal_flags : Marshal.extern_flags list
end

module ER_Diagram (DG: DiGraph)
: ER_DiGraph
=
struct
	type name =
		string
	type relation =
		name
	type concept =
		| Concept  of name
		| ConceptN of name * name
		| ConceptI of name * int
		| ConceptS of name * string 
		| ConceptD of marked (* marker definition *)
		| ConceptU of marked (* marker usage *)
		| ConceptV of marked (* variable definition *)
	and marked =
		{mutable concept:concept; marker:name} 
	type pattern =
		(concept * relation * concept) list
	type t =
		(concept,relation) DG.t
	let make =
		DG.make
	let insert = 
		DG.insert
	let success =
		ref (fun x -> x)
	let concept = function 
		| ConceptD m | ConceptU m -> m.concept
		| c -> c
	let rec select pat t =
		let range1 tl v c =
			v.concept <- c; select tl t
		and range2 tl va vb ca cb =
			va.concept <- ca; vb.concept <- cb; select tl t
		in match pat with
		| [] -> !success ()
		| (ConceptV va,r,ConceptV vb)::tl ->
			DG.find_from_to r (range2 tl va vb) t
		| (ConceptV v,r,c)::tl ->
			DG.find_from (range1 tl v) r (concept c) t
		| (c,r,ConceptV v)::tl ->
			DG.find_to (concept c) r (range1 tl v) t
		| (ca,r,cb)::tl ->
			if DG.member (concept ca) r (concept cb) t
			then select tl t
	let empty =
		Concept ""
	let marshal_flags =
		DG.marshal_flags
end

module Diagram
: DiGraph
=
struct
	type ('a,'b) t =
		{mutable content: ('b,'a BiMap.t) Dictionary.t}
	let make () =
		{content=Dictionary.empty}
	let insert ca r cb t = 
		t.content <- Dictionary.insert
		r (BiMap.singleton ca cb) (BiMap.insert ca cb) t.content
	let member ca r cb t =
		Dictionary.member r (BiMap.member ca cb) t.content 
	let find_from_to r f t =  
		Dictionary.iterate r (BiMap.iterate f) t.content 
	let find_from f r cb t = 
		Dictionary.iterate r (BiMap.iterate_fst cb f) t.content 
	let find_to ca r f t = 
		Dictionary.iterate r (BiMap.iterate_snd ca f) t.content 
	let marshal_flags =
		[]
end
